{% include 'version_description' %}
<table>
  <thead>
    <tr class="header">
      <th style="text-align: left;">Package</th>
      <th style="text-align: left;">Version</th>
    </tr>
  </thead>
  <tbody>
{% for pkg_ver in report.raw_report.versions %}
  <tr>
    <td style="text-align: left;">{{ pkg_ver[0] }}</td>
    <td style="text-align: left;">{{ pkg_ver[1] }}</td>
  </tr>
{% endfor %}
  </tbody>
</table>
