{% include 'metrics_meta_table_text' %}
<table class="table table-striped table-borderless metrics-meta-table" style="width:100%; ">
    <caption style="caption-side: top; text-align: center;">
    Table of metrics metadata
    </caption>
    <colgroup>
    <col style="width: 40%">
    <col style="width: 20%">
    <col style="width: 20%">
    <col style="width: 20%">
    </colgroup>
    <thead style="text-align: left;">
    <tr class="header" style="text-align: left;">
        <th colspan="1">Name</th>
        <th colspan="1">Normalization</th>
        <th colspan="1">Deadband (%)</th>
        <th colspan="1">Cost Parameters</th>
    </tr>
    </thead>
    <tbody style="text-align: left; overflow-wrap: break-word;">
    {% for pfxobs in report.raw_report.processed_forecasts_observations %}
    <tr>
        <td>{{ pfxobs.name }}</td>
        <td>
        {{ pfxobs.normalization_factor }}
        </td>
        <td>
        {{ pfxobs.uncertainty }}
        </td>
        <td>
        <a href="#cost-{{ pfxobs.cost.name | figure_name_filter }}">{{ pfxobs.cost.name }}</a>
        </td>
    </tr>
    {% endfor %}
    </tbody>
</table>
