<meta charset="utf-8">
<!-- Bootstrap core CSS -->
<link href="https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/bootstrap.min.css" type="text/css" rel="stylesheet">
{% if plotly_version %}
{# plotly js and the python plotly library do not have matching versions #}
{% include "insert_plots.html" %}
<script src="https://cdn.plot.ly/plotly-1.52.3.min.js"></script>
<script src="https://code.jquery.com/jquery-3.4.1.min.js"></script>
<script src="https://maxcdn.bootstrapcdn.com/bootstrap/4.1.3/js/bootstrap.min.js"></script>
{% endif %}
